import hashlib
import logging
from typing import List

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group
from swh.model.swhids import ExtendedSWHID

from .inventory import make_inventory
from .removable import mark_removable

logger = logging.getLogger(__name__)


class SwhidOrUrlParamType(click.ParamType):
    name = "swhid or origin URL"

    def convert(self, value, param, ctx):
        from swh.model.exceptions import ValidationError
        from swh.model.swhids import ExtendedSWHID

        if value.startswith("swh:1:"):
            try:
                return ExtendedSWHID.from_string(value)
            except ValidationError:
                self.fail(f"expected extended SWHID, got {value!r}", param, ctx)
        else:
            sha1 = hashlib.sha1(value.encode("utf-8")).hexdigest()
            swhid = ExtendedSWHID.from_string(f"swh:1:ori:{sha1}")
            logger.info(
                "Assuming “%s” is an origin URL; computed origin swhid: %s",
                value,
                swhid,
            )
            return swhid


DEFAULT_CONFIG = {
    "storage": {
        "cls": "postgresql",
        "db": "dbname=softwareheritage host=db.internal.softwareheritage.org user=guest",
        "objstorage": {
            "cls": "memory",
        },
    },
    "graph_client": {
        "url": "http://granet.internal.softwareheritage.org:5009",
        # timeout is in seconds
        # see https://requests.readthedocs.io/en/latest/user/advanced/#timeouts
        "timeout": 10,
    },
}


@swh_cli_group.group(name="alter", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def alter_cli_group(ctx):
    """Archive alteration tools.

    Location of the configuration should be specified through the environment
    variable ``SWH_CONFIG_FILENAME``.

    Expected config format:

        \b
        storage:
          cls: postgresql
          db: "service=…"
          objstorage:
            cls: memory
        \b
        graph:
          cls: remote
          url: "http://granet.internal.softwareheritage.org:5009"
    """
    from swh.core import config

    conf = config.load_from_envvar(default_config=DEFAULT_CONFIG)
    ctx.ensure_object(dict)
    ctx.obj["config"] = conf
    return ctx


@alter_cli_group.command()
@click.option(
    "--dry-run", is_flag=True, help="perform a trial run with no changes made"
)
@click.option(
    "--output-inventory-subgraph",
    type=click.File(mode="w", atomic=True),
)
@click.option(
    "--output-removable-subgraph",
    type=click.File(mode="w", atomic=True),
)
@click.option(
    "--output-pruned-removable-subgraph",
    type=click.File(mode="w", atomic=True),
)
@click.argument(
    "swhids",
    metavar="<SWHID|URL>..",
    type=SwhidOrUrlParamType(),
    required=True,
    nargs=-1,
)
@click.pass_obj
def remove(
    obj,
    swhids: List[ExtendedSWHID],
    dry_run: bool,
    output_inventory_subgraph,
    output_removable_subgraph,
    output_pruned_removable_subgraph,
):
    """Remove the given SWHIDs or URLs from the archive."""
    conf = obj["config"]
    from swh.storage import get_storage

    storage = get_storage(**conf["storage"])
    from swh.graph.http_client import RemoteGraphClient

    graph_client = RemoteGraphClient(**conf["graph_client"])

    click.echo("Removing the following origins:")
    for swhid in swhids:
        click.echo(f" - {swhid}")
    inventory_subgraph = make_inventory(storage, graph_client, swhids)
    if output_inventory_subgraph:
        inventory_subgraph.write_dot(output_inventory_subgraph)
        output_inventory_subgraph.close()
    removable_subgraph = mark_removable(storage, graph_client, inventory_subgraph)
    if output_removable_subgraph:
        removable_subgraph.write_dot(output_removable_subgraph)
        output_removable_subgraph.close()
    removable_subgraph.delete_unremovable()
    if output_pruned_removable_subgraph:
        removable_subgraph.write_dot(output_pruned_removable_subgraph)
        output_pruned_removable_subgraph.close()
    if dry_run:
        removable_swhids = list(removable_subgraph.removable_swhids())
        click.echo(f"We would remove {len(removable_swhids)} objects:")
        for swhid in removable_swhids:
            click.echo(f" - {swhid}")
        return
    raise NotImplementedError("Actual removal still need to be written")



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import shutil
import subprocess

from click.testing import CliRunner
import pytest

from ..cli import DEFAULT_CONFIG, remove
from .test_inventory import (  # noqa
    directory_6_with_multiple_entries_pointing_to_the_same_content,
    snapshot_20_with_multiple_branches_pointing_to_the_same_head,
)
from .test_inventory import graph_client_with_only_initial_origin  # noqa: F401
from .test_inventory import origin_with_submodule  # noqa: F401
from .test_inventory import sample_populated_storage  # noqa: F401
from .test_removable import inventory_from_forked_origin  # noqa: F401
from .test_removable import storage_with_references_from_forked_origin  # noqa: F401


@pytest.fixture
def mocked_external_resources(
    mocker,
    graph_client_with_only_initial_origin,  # noqa: F811
    storage_with_references_from_forked_origin,  # noqa: F811
):
    mocker.patch(
        "swh.storage.get_storage",
        return_value=storage_with_references_from_forked_origin,
    )
    mocker.patch(
        "swh.graph.http_client.RemoteGraphClient",
        return_value=graph_client_with_only_initial_origin,
    )


def test_remove_dry_run_with_origin_as_swhid(mocked_external_resources):
    runner = CliRunner()
    result = runner.invoke(
        remove,
        ["--dry-run", "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165"],
        obj={"config": DEFAULT_CONFIG},
    )
    assert result.exit_code == 0
    assert "We would remove 11 objects" in result.output


def test_remove_dry_run_with_origin_as_url(mocked_external_resources):
    runner = CliRunner()
    result = runner.invoke(
        remove,
        ["--dry-run", "https://example.com/swh/graph2"],
        obj={"config": DEFAULT_CONFIG},
    )
    assert result.exit_code == 0
    assert "We would remove 11 objects" in result.output


def test_remove_dry_run_with_multiple_origins(mocked_external_resources):
    runner = CliRunner()
    result = runner.invoke(
        remove,
        [
            "--dry-run",
            "https://example.com/swh/graph",
            "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
        ],
        obj={"config": DEFAULT_CONFIG},
    )
    assert result.exit_code == 0
    assert "We would remove 23 objects" in result.output


@pytest.mark.skipif(
    not shutil.which("gc"), reason="missing `gc` executable from graphviz"
)
def test_remove_output_inventory_subgraph(mocked_external_resources):
    runner = CliRunner()
    with runner.isolated_filesystem():
        runner.invoke(
            remove,
            [
                "--dry-run",
                "--output-inventory-subgraph=subgraph.dot",
                "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
            ],
            obj={"config": DEFAULT_CONFIG},
        )
        completed_process = subprocess.run(
            ["gc", "subgraph.dot"], check=True, capture_output=True
        )
        assert b"      21      24 Inventory" in completed_process.stdout


@pytest.mark.skipif(
    not shutil.which("gc"), reason="missing `gc` executable from graphviz"
)
def test_remove_output_removable_subgraph(mocked_external_resources):
    runner = CliRunner()
    with runner.isolated_filesystem():
        runner.invoke(
            remove,
            [
                "--dry-run",
                "--output-removable-subgraph=subgraph.dot",
                "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
            ],
            obj={"config": DEFAULT_CONFIG},
        )
        completed_process = subprocess.run(
            ["gc", "subgraph.dot"], check=True, capture_output=True
        )
        assert b"      21      24 Removable" in completed_process.stdout


@pytest.mark.skipif(
    not shutil.which("gc"), reason="missing `gc` executable from graphviz"
)
def test_remove_output_pruned_removable_subgraph(mocked_external_resources):
    runner = CliRunner()
    with runner.isolated_filesystem():
        runner.invoke(
            remove,
            [
                "--dry-run",
                "--output-pruned-removable-subgraph=subgraph.dot",
                "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
            ],
            obj={"config": DEFAULT_CONFIG},
        )
        completed_process = subprocess.run(
            ["gc", "subgraph.dot"], check=True, capture_output=True
        )
        assert b"      11      10 Removable" in completed_process.stdout
